/* David Leonard, 2002. Public domain. */
/*
 * Some small fragments of code is from OpenBSD's ftp/util.c credited to
 * Enami Tsugutomo <enami@ba2.so-net.or.jp>, Todd Miller <millert@openbsd.org>
 * and Luke Mewburn <lukem@openbsd.org>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <ctype.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include "progressbar.h"

#define DEFAULTHIDE	4
#define DEFAULTCONFIG	"4 ftp 0 scp 0"

#define TITLELEN	16	/* maximum title length */
#define INFOLEN		19	/* percentage + eta + padding */

static volatile int resize_needed = 0;		/* sigwinch received */

struct progressbar {
	FILE *file;
	int ttywidth;
	unsigned long max, pos;
	int hide;
	int flags;
#define PB_FORCE	0x10000
#define PB_DRAWN	0x20000
#define PB_DONE		0x40000
	struct timeval started;
	int lasteta, lastetavalid, laststarpos, lastpc;
	char title[TITLELEN + 1];
};

/* Use TIOCGWINSZ to get the screen size */
static int
getttywidth(ttyfile)
	FILE *ttyfile;
{
	struct winsize winsize;

	if (ioctl(fileno(ttyfile), TIOCGWINSZ, &winsize) != -1)
		return winsize.ws_col ? winsize.ws_col : 80;
	else
		return 80;
}

/* Return true if this is a foreground process */
static int
foregroundproc()
{
	static pid_t pgrp = -1;
	int ctty_pgrp;

	if (pgrp == -1)
		pgrp = getpgrp();
	return((ioctl(STDOUT_FILENO, TIOCGPGRP, &ctty_pgrp) != -1 &&
	    ctty_pgrp == pgrp));
}

/*
 * Return the number of seconds to wait before displaying the progress
 * bar. If 0 is returned, the bar should always be displayed. If -1
 * is returned, then the bar should never be displayed.
 */
static int
gethide()
{
	const char *ev, *n;
	int match, defaulthide = DEFAULTHIDE;
	long v;
	extern char *__progname;
	int prognamelen = strlen(__progname);

	/*
	 * Valid examples of $PROGRESSBAR are:
	 *	"7"		(default 7)
	 *	"-1"		(default -1)
	 *	""		(default 4 (=DEFAULTHIDE))
	 *	"3 ftp 0"	(default 3, use 0 for ftp)
	 *	"ftp 1"		(default 4, use 1 for ftp)
	 *	"ftp 1 cp 8"	(default 4, use 1 for ftp and 8 for cp)
	 */

	ev = getenv("PROGRESSBAR");
	if (ev == NULL)
		ev = DEFAULTCONFIG;

	/* Get default, if it looks like a number */
	v = strtoul(ev, (char **)&n, 0);
	if (ev != n) {
		defaulthide = v;
		ev = n;
	}

	while (*ev && isspace(*ev))
		ev++;
	while (*ev) {
		/* Get program name */
		for (n = ev; *n; n++)
			if (isspace(*n))
				break;
		match = n - ev == prognamelen
			&& memcmp(__progname, ev, prognamelen) == 0;
		ev = n;
		v = strtoul(ev, (char **)&n, 0);
		if (ev == n)
			break;		/* Problem */
		if (match)
			return v;
		ev = n;
		while (*ev && isspace(*ev))
			ev++;
	}

	return defaulthide;
}

/* Allocate and draw initial progress bar */
struct	progressbar *
progressbar_start(file, max, title, flags)
	FILE *file;
	unsigned long max;
	const char *title;
	int flags;
{
	struct progressbar *p;

	p = (struct progressbar *)malloc(sizeof *p);
	if (p == NULL)
		return NULL;

	p->file = file;
	p->flags = flags & ~(PB_DRAWN|PB_FORCE|PB_DONE);
	if (!isatty(fileno(file)))
		p->hide = -1;		/* never show on non-tty */
	else
		p->hide = gethide();
	p->ttywidth = getttywidth(file);
	p->pos = 0;
	gettimeofday(&p->started, NULL);
	if ((flags & PB_NOWINCH) == 0)
		signal(SIGWINCH, progressbar_winch);
	progressbar_change(p, max, title, flags);
	return p;
}

/* Alter parameters and re-draw progress bar */
void
progressbar_change(p, max, title, flags)
	struct	progressbar *p;
	unsigned long max;
	const char *title;
	int flags;
{
	p->max = max;
	if (title)
		strlcpy(p->title, title, sizeof p->title);
	else
		p->title[0] = '\0';
	p->flags |= PB_FORCE;
	progressbar_update(p, p->pos);
}

/* Update progress bar, redrawing only if necessary */
void
progressbar_update(p, pos)
	struct progressbar *p;
	unsigned long pos;
{
	int i, maxstars, starpos, pc, etavalid;
	unsigned int eta;
	struct timeval now, delta;
	gettimeofday(&now, NULL);
	timersub(&now, &p->started, &delta);

	/* Have progress bars been suppressed? */
	if (p->hide < 0
	    || (p->hide > 0 && delta.tv_sec < p->hide))
		return;

	/* Are we in the background */
	if (!foregroundproc()) {
		/* Immediate redraw when we are next foregrounded */
		p->flags = PB_FORCE;
		return;
	}

	/* Check for a window resize */
	if (resize_needed) {
		resize_needed = 0;
		p->ttywidth = getttywidth(p->file);
		p->flags |= PB_FORCE;
	}

	/* Compute number of stars to draw */
	maxstars = p->ttywidth - INFOLEN;
	if (p->title[0] != '\0')
		maxstars -= TITLELEN;
	if (maxstars < 0)
		maxstars = 0;
	if (p->max != 0)
		starpos = maxstars * pos / p->max;
	else
		starpos = 0;
	if (starpos > maxstars)
		starpos = maxstars;

	/* Compute percentage complete */
	if (p->max == 0)
		pc = 100;
	else
		pc = 100 * pos / p->max;

	/* Compute time remaining */
	if (pos >= p->max) {
		eta = 0;
		etavalid = 1;
	} else if (pos != 0) {
		eta = 1 + (p->max - pos) 
		    * (delta.tv_sec * 1000000 + delta.tv_usec)
		    / pos / 1000000;
		etavalid = 1;
	} else
		etavalid = 0;

	/* Don't redraw if the numbers haven't changed */
	if ((p->flags & PB_FORCE) == 0
	    && p->laststarpos == starpos
	    && p->lastpc == pc
	    && p->lastetavalid == etavalid
	    && (!etavalid || p->lasteta == eta))
		return;

	if ((p->flags & PB_DRAWN) != 0)
		fprintf(p->file, "\r");
	if (p->title[0] != '\0')
		fprintf(p->file, "%-*.*s", TITLELEN, TITLELEN, p->title);
	for (i = 0; i < maxstars; i++)
		if (i < starpos)
			fprintf(p->file, "*");
		else
			fprintf(p->file, " ");
	fprintf(p->file, " %3d%%", pc);
	if ((p->flags & PB_DONE) != 0) {
	    int elap = delta.tv_sec;
	    fprintf(p->file, "    %3d:%02d:%02d", 
		elap / 3600, (elap / 60) % 60, elap % 60);
	} else if (etavalid) 
	    fprintf(p->file, " ETA%3d:%02d:%02d", 
		eta / 3600, (eta / 60) % 60, eta % 60);
	else
	    fprintf(p->file, " ETA --:--:--");

	fflush(p->file);

	/* Remember numbers for later */
	p->pos = pos;
	p->laststarpos = starpos;
	p->lastpc = pc;
	p->lastetavalid = etavalid;
	memcpy(&p->lasteta, &eta, sizeof p->lasteta);
	p->flags &= ~PB_FORCE;
	p->flags |= PB_DRAWN;
}

/* Deallocate progress bar. */
void
progressbar_finish(p)
	struct progressbar *p;
{
	p->flags |= PB_DONE;
	if ((p->flags & PB_DRAWN) != 0) {
		p->flags |= PB_FORCE;
		progressbar_update(p, p->pos);
		fprintf(p->file, "\n");
		fflush(p->file);
	}
	free(p);
}

/* Handle SIGWINCH signals */
void
progressbar_winch(sig)
	int sig;
{
	resize_needed = 1;
}
